"""
# usage:
>>> from .improved_invoke import improved_task as task # noqa
>>> @task(flags={})
>>> def something(): ...
"""

import inspect
import typing
import warnings
from typing import Any, Callable, Iterable, List, Optional

from docstring_parser import parse as parse_docstring
from fabric import Connection
from invoke import Argument
from invoke.context import Context
from invoke.tasks import Task as InvokeTask
from invoke.tasks import task as invoke_task
from typing_extensions import Unpack

from .helpers import AnyDict

TaskFn: typing.TypeAlias = typing.Callable[[Context], Any] | typing.Callable[..., Any]

P = typing.ParamSpec("P")
R = typing.TypeVar("R")


def extract_arg_doc(docstring: str, arg_name: str):
    doc = parse_docstring(docstring)
    for param in doc.params:
        if param.arg_name != arg_name:
            continue

        return param.description

    return None


class TaskOptions(typing.TypedDict, total=False):
    name: Optional[str]
    aliases: Iterable[str]
    positional: Optional[Iterable[str]]
    optional: Iterable[str]
    default: bool
    auto_shortflags: bool
    help: Optional[AnyDict]
    pre: Optional[list[TaskFn]]
    post: Optional[list[TaskFn]]
    autoprint: bool
    iterable: Optional[Iterable[str]]
    incrementable: Optional[Iterable[str]]
    flags: dict[str, Iterable[str]] | None
    hookable: Optional[bool]


class TaskCallable(typing.Protocol):
    def __call__(
        self, **_: Unpack[TaskOptions]
    ) -> Callable[
        [Callable[P, R]],
        Callable[P, R],
    ]: ...


class ImprovedTask(InvokeTask[TaskCallable]):
    """
    Improved version of Invoke Task where you can set custom flags for command line arguments.
    This allows you to specify aliases, rename (e.g. --json for 'as_json')  and custom short flags (--exclude = -x)
    """

    _flags: dict[str, list[str]]

    def __init__(
        self,
        body: TaskCallable,
        name: Optional[str] = None,
        aliases: Iterable[str] = (),
        positional: Optional[Iterable[str]] = None,
        optional: Iterable[str] = (),
        default: bool = False,
        auto_shortflags: bool = True,
        help: Optional[AnyDict] = None,  # noqa
        pre: Optional[list[TaskFn]] = None,
        post: Optional[list[TaskFn]] = None,
        autoprint: bool = False,
        iterable: Optional[Iterable[str]] = None,
        incrementable: Optional[Iterable[str]] = None,
        # new:
        flags: dict[str, Iterable[str]] | None = None,
        hookable: Optional[bool] = None,
    ):
        self._flags = flags or {}
        self.hookable = hookable

        super().__init__(
            body=body,
            name=name,
            aliases=tuple(aliases),
            positional=positional,
            optional=optional,
            default=default,
            auto_shortflags=auto_shortflags,
            help=help,
            pre=pre,  # type: ignore
            post=post,  # type: ignore
            autoprint=autoprint,
            iterable=iterable,
            incrementable=incrementable,
        )

    def arg_opts(self, name: str, default: str, taken_names: Iterable[str]) -> AnyDict:
        """Get argument options.

        Args:
            name (str): The name of the argument.
            default (str): The default value of the argument.
            taken_names (Iterable[str]): Names that have already been taken.

        Returns:
            AnyDict: A dictionary of argument options.
        """
        opts = super().arg_opts(name=name, default=default, taken_names=set(taken_names))
        help_str = self.help.pop(name, None) or opts.get("help") or extract_arg_doc(self.body.__doc__, name)
        opts["help"] = opts.get("help") or help_str

        if flags := self._flags.get(name):
            opts["names"] = list(flags)
            for synonym in list(flags):
                self.help[synonym] = help_str

        return opts

    def get_arguments(self, ignore_unknown_help: Optional[bool] = None) -> List[Argument]:
        return super().get_arguments(ignore_unknown_help=True)

    def _execute_subtask(self, ctx: Context, task: TaskFn, *args, **kwargs):
        """Execute a subtask with provided context and arguments.

        Args:
            ctx (Context): The context to pass to the task.
            task (TaskFn): The task function to execute.
            *args: Positional arguments for the task.
            **kwargs: Keyword arguments for the task.
        """
        sig = inspect.signature(task)
        task_args = [ctx]  # Start with the context
        task_kwargs = {}

        # Collect positional arguments
        param_names = list(sig.parameters.keys())
        for i, param in enumerate(param_names[1:], start=0):  # Skip 'ctx'
            if i < len(args):
                task_args.append(args[i])
            elif sig.parameters[param].default is sig.empty:
                raise ValueError(f"Missing required argument: {param}")

        # Collect keyword arguments
        for param in param_names[1:]:
            if param in kwargs:
                task_kwargs[param] = kwargs[param]

        # Call the task with the prepared arguments
        return task(*task_args, **task_kwargs)

    def _run_hooks(self, ctx: Context, *args, **kwargs):
        """Run hooks for the current instance.

        Args:
            ctx (Context): The context to pass to the hooks.
            *args: Positional arguments for the hooks.
            **kwargs: Keyword arguments for the hooks.
        """
        for namespace, task in find_task_across_namespaces(self.name).items():
            if task is not self and getattr(task, "hookable", False) is not False:
                try:
                    subresult = self._execute_subtask(ctx, task, *args, **kwargs)
                except Exception as e:
                    warnings.warn(
                        f"Failed running subtask {namespace}.{task.name}: {e}.", source=e, category=RuntimeWarning
                    )
                    continue

                if isinstance(ctx["result"], dict) and isinstance(subresult, dict):
                    ctx["result"].update(subresult)
                elif subresult is not None:
                    ctx["result"] = subresult

    def __call__(self, ctx: Context | Connection, *args, **kwargs):
        """Invoke the callable instance.

        Args:
            ctx (Context): The context to pass.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            The result of the superclass call.
        """
        # ctx.get works for Context but not for Connection!
        setattr(ctx, "result", getattr(ctx, "result", {}))  # ctx["result"] = ctx.get("result") or {}

        result = super().__call__(ctx, *args, **kwargs)

        if isinstance(ctx["result"], dict) and isinstance(result, dict):
            ctx["result"].update(result)
        elif result is not None:
            ctx["result"] = result

        if self.hookable:
            self._run_hooks(ctx, *args, **kwargs)

        return ctx["result"]


def find_task_across_namespaces(name: str) -> dict[str, ImprovedTask]:
    from .cli import collection

    return {ns.name: task for ns in collection.collections.values() if (task := ns.tasks.get(name))}


def improved_task(*fn: Optional[TaskCallable], **options: Unpack[TaskOptions]) -> TaskCallable:
    """
    Marks wrapped callable object as a valid Invoke task.

    This function may be called without any parentheses if no extra options need to be
    specified. Otherwise, the following keyword arguments are allowed in the
    parenthesized form:

    Args:
        name (str): Default name to use when binding to a `.Collection`. Useful for
            avoiding Python namespace issues (i.e. when the desired CLI level name
            can't or shouldn't be used as the Python level name.)
        aliases (List[str]): Specify one or more aliases for this task, allowing it to be
            invoked as multiple different names. For example, a task named ``mytask``
            with a simple ``@task`` wrapper may only be invoked as ``"mytask"``.
            Changing the decorator to be ``@task(aliases=['myothertask'])`` allows
            invocation as ``"mytask"`` *or* ``"myothertask"``.
        positional (Iterable[str]): Iterable overriding the parser's automatic "args with no
            default value are considered positional" behavior. If a list of arg
            names, no args besides those named in this iterable will be considered
            positional. (This means that an empty list will force all arguments to be
            given as explicit flags.)
        optional (Iterable[str]): Iterable of argument names, declaring those args to
            have optional values. Such arguments may be
            given as value-taking options (e.g. ``--my-arg=myvalue``, wherein the
            task is given ``"myvalue"``) or as Boolean flags (``--my-arg``, resulting
            in ``True``).
        iterable (Iterable[str]): Iterable of argument names, declaring them to build
            iterable values.
        incrementable (Iterable[str]): Iterable of argument names, declaring them to
            increment their values.
        default (bool): Boolean option specifying whether this task should be its
            collection's default task (i.e. called if the collection's own name is
            given.)
        auto_shortflags (bool): Whether or not to automatically create short
            flags from task options; defaults to True.
        help (Dict[str, str]): Dict mapping argument names to their help strings. Will be
            displayed in ``--help`` output. For arguments containing underscores
            (which are transformed into dashes on the CLI by default), either the
            dashed or underscored version may be supplied here.
        pre (List[TaskCallable]): Lists of task objects to execute prior to the wrapped
            task whenever it is executed.
        post (List[TaskCallable]): Lists of task objects to execute after the wrapped
            task whenever it is executed.
        autoprint (bool): Boolean determining whether to automatically print this
            task's return value to standard output when invoked directly via the CLI.
            Defaults to False.
        flags (dict[str, list[str]]): Mapping of flag names that modify task behavior.
                               e.g. `@task(flags={'exclude': ['--exclude', '-x'], 'as_json': ['--json']})`
        hookable (Optional[bool]): Boolean option that controls whether the task can be hooked by plugins.
                                       - **True**: This setting is primarily used by core tasks. It allows the task to look for other tasks (from plugins or local definitions) with the same name and execute them after the main task completes. This enables a cascading execution of tasks, enhancing modularity and reusability.
                                       - **False**: This setting is typically used by local or plugin tasks to indicate that they do not want to be hooked by core tasks, even if they share the same name. This ensures that the task remains isolated and does not trigger any unintended behavior from cascading executions.
                                       - **None** (default): This represents the default behavior. For core tasks, it means that the task will not search for other tasks with the same name to execute. For local or plugin tasks, it allows them to be hooked by other tasks.
        fn (TaskCallable): when you use `@task` without parentheses, this is the function you're decorating.
                            Using `@task()` with parens is recommended for better type-hints.

    If any non-keyword arguments are given, they are taken as the value of the
    ``pre`` kwarg for convenience's sake. (It is an error to give both
    ``*args`` and ``pre`` at the same time.)
    """
    return invoke_task(*fn, **options, klass=ImprovedTask)


__all__ = ["ImprovedTask", "improved_task"]



"""
This file exposes some functions so this tool can be used as a library.
"""

# SPDX-FileCopyrightText: 2023-present Remco Boerma <remco.b@educationwarehouse.nl>
#
# SPDX-License-Identifier: MIT

import warnings

from cryptography.utils import CryptographyDeprecationWarning

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

from . import tasks  # noqa E402 - import has to come after warning filter
from .constants import DOCKER_COMPOSE  # noqa E402  - import has to come after warning filter
from .helpers import (  # noqa E402  - import has to come after warning filter
    KEY_ARROWDOWN,
    KEY_ARROWUP,
    KEY_ENTER,
    AnyDict,
    Logger,
    NoopLogger,
    VerboseLogger,
    add_alias,
    arg_was_passed,
    confirm,
    dc_config,
    dump_set_as_list,
    executes_correctly,
    execution_fails,
    fabric_read,
    fabric_read_bytes,
    fabric_write,
    flatten,
    generate_password,
    interactive_selected_checkbox_values,
    interactive_selected_radio_value,
    kwargs_to_options,
    noop,
    print_aligned,
    print_box,
    shorten,
    yaml_loads,
)
from .meta import is_installed  # noqa E402  - import has to come after warning filter
from .health import (
    HealthLevel,
    get_healths,
    find_container_ids,
    find_containers_ids,
    HealthStatus,
    inspect,
)  # noqa E402  - import has to come after warning filter
from .improved_invoke import ImprovedTask, improved_task  # noqa E402  - import has to come after warning filter
from .tasks import (  # noqa E402  - import has to come after warning filter
    TomlConfig,
    check_env,
    get_env_value,
    get_task,
    read_dotenv,
    set_env_value,
    task_for_namespace,
)

Task = ImprovedTask
task = improved_task

__all__ = [
    "DOCKER_COMPOSE",
    "KEY_ARROWDOWN",
    "KEY_ARROWUP",
    "KEY_ENTER",
    "AnyDict",
    "ImprovedTask",
    "Logger",
    "NoopLogger",
    "Task",
    "TomlConfig",
    "VerboseLogger",
    "add_alias",
    "arg_was_passed",
    "check_env",
    "confirm",
    "dc_config",
    "dump_set_as_list",
    "executes_correctly",
    "execution_fails",
    "fabric_read",
    "fabric_read_bytes",
    "fabric_write",
    "flatten",
    "generate_password",
    "get_env_value",
    "get_task",
    "improved_task",
    "interactive_selected_checkbox_values",
    "interactive_selected_radio_value",
    "kwargs_to_options",
    "noop",
    "print_aligned",
    "print_box",
    "read_dotenv",
    "set_env_value",
    "shorten",
    "task",
    "task_for_namespace",
    "tasks",
    "yaml_loads",
    "is_installed",
    "HealthLevel",
    "get_healths",
    "find_container_ids",
    "find_containers_ids",
    "HealthStatus",
    "inspect",
]



import datetime as dt
import itertools
import json
import re
import sys
import typing
from datetime import datetime
from pathlib import Path
from typing import Optional

import anyio
from termcolor import cprint

""" --- translated from colors.go in docker compose """


def ansi_color_code(code: str, format_opts: typing.Collection[str] = ()) -> str:
    res = "\033["
    for c in format_opts:
        res += f"{c};"
    return f"{res}{code}m"


ColorFn = typing.Callable[[str], str]


def make_color_func(code: str) -> ColorFn:
    return lambda s: f"{ansi_color_code(code)}{s}{ansi_color_code('0')}"


def build_rainbow() -> tuple[ColorFn, ...]:
    names = (
        "grey",
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
    )

    colors = {}
    for i, name in enumerate(names):
        colors[name] = make_color_func(str(30 + i))
        colors[f"intense_{name}"] = make_color_func(f"{30 + i};1")

    return (
        colors["cyan"],
        colors["yellow"],
        colors["green"],
        colors["magenta"],
        colors["blue"],
        colors["intense_cyan"],
        colors["intense_yellow"],
        colors["intense_green"],
        colors["intense_magenta"],
        colors["intense_blue"],
    )


def rainbow() -> typing.Generator[str, None, None]:
    """
    rainbow = []colorFunc{
                colors["cyan"],
                colors["yellow"],
                colors["green"],
                colors["magenta"],
                colors["blue"],
                colors["intense_cyan"],
                colors["intense_yellow"],
                colors["intense_green"],
                colors["intense_magenta"],
                colors["intense_blue"],
        }

    Yield colors from the docker compose rainbow map in a cyclic way.
    """
    yield from itertools.cycle(build_rainbow())


""" -- end of colors.go """

FilterFn: typing.TypeAlias = Optional[typing.Callable[[str], bool]]


async def parse_docker_log_line(
    line: str,
    human_name: str,
    container_id: str,  # noqa ARG001 - may be used later
    color: ColorFn,
    stream: Optional[str] = None,
    since: Optional[str] = None,
    re_filter: FilterFn = None,
    show_ts: bool = True,  # full, default, no
    verbose: bool = False,
) -> bool:
    """
    Parses a single line of Docker log.

    Args:
        line (str): The log line to parse as a JSON string.
        human_name (str): A human-readable name for the container.
        container_id (str): The ID of the container generating the log.
        color (ColorFn): A function to apply color formatting to the output.
        stream (Optional[str]): The log stream type (stdout/stderr) to filter by.
        since (Optional[str]): A timestamp to filter logs that are earlier than this value.
        re_filter (FilterFn): A regex filter function to apply to the log message.
        show_ts (bool): If true, include the timestamp in the output.
        verbose (bool): If true, show detailed timestamps.

    Returns:
        bool: True if the log line was processed successfully and something was actually printed, False otherwise.
    """
    try:
        data = json.loads(line)
    except json.JSONDecodeError:
        # don't crash, just ignore
        return False

    # data containers log, stream (stdout/stderr) and time.

    if stream and data["stream"] != stream:
        return False

    if since and data["time"] < since:
        return False

    if not (log := data.get("log")):
        return False

    if re_filter and not re_filter(log):
        return False

    if show_ts:
        # iso is up to 30 chars wide
        timestamp = data["time"].ljust(30, " ") if verbose else data["time"].split(".")[0]

        prefix = color(f"{human_name} |") + f" {timestamp} |"
    else:
        prefix = color(f"{human_name} |")

    print(prefix, log, end="")
    return True


TD_RE = re.compile(r"(\d+)\s*(h(our)?|m(inute)?|s(econd)?|d(ay)?)s?\s*(ago)?", re.IGNORECASE)


def _parse_timedelta(since: str) -> Optional[dt.timedelta]:
    if since == "now":
        # 0.1s because otherwise the timedelta is falsey
        return dt.timedelta(seconds=0.1)

    if not (since and (match := TD_RE.match(since))):
        return None

    number = int(match.group(1))
    unit = match.group(2)

    match unit.lower():
        case "day" | "d":
            return dt.timedelta(days=number)
        case "hour" | "h":
            return dt.timedelta(hours=number)
        case "minute" | "m":
            return dt.timedelta(minutes=number)
        case "second" | "s":
            return dt.timedelta(seconds=number)
        case _:
            print(f"Warn: unrecognized unit {unit}.", file=sys.stderr)
            return None


def utcnow() -> dt.datetime:
    """
    Replacement of datetime.utcnow.

    When 3.13 is released, 3.10 can be dropped and this check can also be removed.
    """
    if sys.version_info < (3, 11):
        # deprecated since 3.12
        return datetime.utcnow()
    else:
        # dt.UTC doesn't exist yet in 3.10
        return datetime.now(dt.UTC)


def parse_timedelta(since: str, utc: bool = True) -> str:
    # turns human-readable 'since' into a iso datetime string.

    if delta := _parse_timedelta(since):
        now = utcnow() if utc else datetime.now()
        return (now - delta).isoformat()

    return since


POSSIBLE_FLAGS = {
    # https://docs.python.org/3/library/re.html
    "a": re.ASCII,
    "d": re.DEBUG,
    "i": re.IGNORECASE,
    "l": re.LOCALE,
    "m": None,  # re.MULTILINE but the logger works line-by-line so this isn't really possible
    "s": re.DOTALL,
    "u": re.UNICODE,
    "x": re.VERBOSE,
    # custom: 'v' to invert
}


def parse_regex(raw: str) -> FilterFn:
    """
    Turn `/pattern/flags` into a Regex object.

    Uses the grep style flags (i for case insensitive, v for invert)
    """

    # zero slashes: just a pattern, no flags.
    # one slash: search term with / in it
    # two slashes (+ starts with /): regex with flags
    # more slashes: flags AND / in filter itself

    if raw.startswith("/") and raw.count("/") > 1:
        # flag-mode
        _, *rest, flags_str = raw.split("/")
        flags = set(flags_str.lower())
        pattern = "/".join(rest)
    else:
        # normal search mode, no flags
        flags = set()
        pattern = raw

    flags_bin = 0  # re.NOFLAG doesn't exist in 3.10 yet

    for flag in flags:
        flags_bin |= POSSIBLE_FLAGS.get(flag) or 0  # re.NOFLAG

    re_compiled = re.compile(pattern, flags_bin)

    if "v" in flags:
        # v for inverse like `grep -v`
        return lambda text: not re_compiled.search(text)
    else:
        return lambda text: bool(re_compiled.search(text))


class TailConfig(typing.TypedDict):
    filename: str
    human_name: str
    container_id: str
    stream: Optional[str]
    since: Optional[str]
    re_filter: Optional[FilterFn]
    color: ColorFn
    timestamps: bool
    verbose: bool
    state: str


async def tail(config: TailConfig) -> None:
    print_args = " ".join(sys.argv[1:])
    fname = config["filename"]
    fpath = Path(fname)

    if not fpath.exists():
        cprint(f"Expected log file but found nothing. Container {config['human_name']} state: {config['state']}")
        return

    async with await anyio.open_file(fname) as f:
        while True:
            # if the file doesn't exist, the container was removed (dc down)
            exited = config["state"] == "exited" or not fpath.exists()

            if contents := await f.readline():
                print(" " * 20, end="\r")
                did_print = await parse_docker_log_line(
                    contents,
                    config["human_name"],
                    config["container_id"],
                    stream=config.get("stream"),
                    since=config.get("since"),
                    re_filter=config.get("re_filter"),
                    color=config["color"],
                    show_ts=config["timestamps"],
                    verbose=config["verbose"],
                )

                if did_print:
                    cprint(f"$ edwh {print_args}", color="white", end="\r")

            elif exited:
                # if state = 'exited' and last line was reached -> stop
                # note: this only happens when the container was already shut down when 'logs' started,
                # exited is NOT updated live. This was done on purpose, so restarting containers keep being logged.
                # but if the container was already shut down, it's probably something like migrate,
                # which will not restart automatically, so you probably don't want to wait forever.
                return
            else:
                # Add a small sleep to reduce CPU load (if no data right now)
                await anyio.sleep(0.1)



"""
This files contains everything to do with meta-tasks such as self-updating
"""

import concurrent.futures
import sys
import typing
from typing import Optional

import yayarl as yarl
from invoke.context import Context
from packaging.version import InvalidVersion, Version
from packaging.version import parse as parse_package_version
from termcolor import cprint

from .helpers import AnyDict
from .improved_invoke import improved_task as task

PYPI_URL_BASE = yarl.URL("https://pypi.python.org/pypi/")


def _python() -> str:
    """
    used to detect current Python environment, even in pipx
    """
    return sys.executable


def _pip(python: str = _python()) -> str:
    """
    used to detect current pip environment, even in pipx
    """
    # uv.find_uv_bin() does not really work here, because then the right venv may not be used!
    return f"{python} -m uv pip"


def _get_pypi_info(package: str) -> AnyDict:
    """
    Load metadata from pypi for a package
    """
    url = PYPI_URL_BASE / package / "json"
    resp = url.get(timeout=10)
    return typing.cast(AnyDict, resp.json())


def _get_latest_version_from_pypi(package: str) -> Version:
    """
    Get the latest Version for a package  from pypi
    """
    data = _get_pypi_info(package)
    if not data or not data.get("info"):
        raise ModuleNotFoundError(f"Plugin {package} does not seem to exist.")

    return parse_package_version(data["info"]["version"])


def _get_available_plugins_from_pypi(package: str, extra: Optional[str] = None) -> list[str]:
    """
    List all plugins available for package, optionally for a specific 'extra'.

    e.g.
    [mypackage]
    dev = ['package1', 'package2']
    another_extra = ['package1', 'package3']

    > _get_available_plugins_from_pypi('mypackage')
    ['package1', 'package2', 'package3']

    > _get_available_plugins_from_pypi('mypackage, 'dev')
    ['package1', 'package2']

    """
    data = _get_pypi_info(package)
    extras = data["info"]["requires_dist"]

    if extra:
        extras = [_.split(";")[0] for _ in extras if _.endswith(f'; extra == "{extra}"')]

    return list(extras)


def _gather_package_metadata_threaded(packages: typing.Iterable[str]) -> dict[str, AnyDict | None]:
    """
    For any package in packages, gather its metadata from pypi
    """
    all_data: dict[str, AnyDict | None] = {}
    with concurrent.futures.ThreadPoolExecutor() as executor:
        pkg_names = [_.split("==")[0] for _ in packages]
        for result, package in zip(executor.map(_get_pypi_info, pkg_names), packages):
            all_data[package] = result

    return all_data


def _determine_newest_version(releases: typing.Collection[str]) -> str:
    sorted_releases = sorted(releases, key=Version)
    return sorted_releases[-1]


def _determine_outdated_threaded(
    installed_plugins: typing.Collection[str], prerelease: bool = False
) -> dict[str, Version]:
    """
    Like _determine_outdated but parallelized with Threading

    installed_plugins is a list (or other iterable) of ["name==version", "name @ location"] type strings
    """
    plugins_metadata = _gather_package_metadata_threaded([_ for _ in installed_plugins if " @ " not in _])

    outdated = {}
    for plugin, metadata in plugins_metadata.items():
        if not metadata:
            continue

        try:
            name, current_version_str = plugin.split("==")
            current_version = parse_package_version(current_version_str)

            latest_stable = metadata["info"]["version"]
            latest_prerelease = _determine_newest_version(metadata["releases"].keys()) if prerelease else None
            latest_version = parse_package_version(latest_prerelease if prerelease else latest_stable)
        except Exception:
            # no current or latest version found? skip
            continue

        if current_version and latest_version and latest_version > current_version:
            outdated[name] = latest_version

    return outdated


def _parse_versions(installed: list[str]) -> dict[str, Version | None]:
    """
    Given a list of installed packages from pip freeze (_plugins), gather the parsed versions
    """
    versions = {}
    for pkg in installed:
        parts = pkg.split(" @ ")[0].split("==")
        name = parts[0]
        try:
            version = parse_package_version(parts[1])
        except (InvalidVersion, IndexError):
            version = None
        # finally:
        versions[name] = version

    return versions


@task()
def plugins(c: Context, verbose: bool = False, changelog: bool = False) -> None:
    """
    alias for plugin.list or plugin.changelog --new
    """
    from .local_tasks import plugin

    if changelog:
        return plugin.changelog(c, [], new=True)
    else:
        return plugin.list_plugins(c, verbose=verbose)


def _self_update(c: Context, prerelease: bool = False, no_cache: bool = False) -> None:
    """
    Wrapper for self-update that can handle type hint Context
    """
    from .local_tasks.plugin import list_installed_plugins

    pip_command = _pip()

    edwh_packages = list_installed_plugins(c, pip_command)
    if not edwh_packages or (len(edwh_packages) == 1 and edwh_packages[0] == ""):
        cprint("No 'edwh' packages found. That can't be right", color="yellow")

    old_plugins = _determine_outdated_threaded(edwh_packages, prerelease=prerelease)

    if not old_plugins:
        cprint("Nothing to update", "blue")
        exit()

    cprint(f"Will try to update {len(old_plugins)} packages.", "blue")

    success = []
    failure = []
    for plugin, version in old_plugins.items():
        command = f"{pip_command} install {plugin}=={version}"
        if no_cache:
            command = f"{command} --no-cache"

        result = c.run(command, warn=True)

        if result and result.return_code == 0:
            success.append(plugin)
        else:
            failure.append(plugin)

    if success:
        cprint(f"{len(success)}/{len(old_plugins)} updated successfully.", "green")

    if failure:
        cprint(f"{', '.join(failure)} failed updating", "red")


@task(
    flags={
        "prerelease": ["prerelease", "pre", "pre-release", "p"],
        "no_cache": ["no-cache", "f", "fresh"],
    }
)
def self_update(c: Context, prerelease: bool = False, no_cache: bool = False) -> None:
    """Updates `edwh` and all installed plugins.

    Args:
        c (Context): invoke ctx
        prerelease (bool, optional): allow non-stable releases? Defaults to False.
        no_cache (bool, optional): download fresh? Defaults to False.
    """
    return _self_update(c, prerelease, no_cache)


def is_installed(ctx: Context, command: str) -> bool:
    """
    Check if a bash command is known.
    """
    return ctx.run(f"which {command}", hide="both", warn=True).ok
